/**
 *   $Date$
 *
 * @short Class representing the readout data structure of the MuCTPI hardware
 *
 * The MuCTPI hardware and the simulation can send all the muon candidates
 * recevied in an up to +-2 BC window together with the muon multiplicities
 * in the same window to the readout system. This class represents the ROD
 * fragment generated by the MuCTPI.
 *
 * For a detailed description of the data format of the MuCTPI, see
 * the MIROD documentation (https://edms.cern.ch/file/248757/1/mirod.pdf).
 *
 * @file CTPfragment/MuCTPIdataformat.h
 * @author <a href="mailto:berge@cern.ch">David Berge</a>
 *
 * $Author$
 * $Revision$
 * 
 */

#ifndef MUCTPI_DATAFORMAT_H
#define MUCTPI_DATAFORMAT_H

#include <inttypes.h>

namespace MuCTPIdataformat {

	/// Enumeration defining the different sources for the muon candidates
	enum SectorLocation { BARREL = 0, ENDCAP = 1, FORWARD = 2 };

	/// Binary 111 representing the maximal multiplicity value for a given threshold
	static const uint32_t MultiplicityVal = 7;

	/// Number of multiplicity bits reserved per threshold
	static const uint32_t MultiplicityBits = 3;

	/// Defining the number of p<sub>T</sub> thresholds in the system
	static const uint32_t MultiplicityThresholdNumber = 6;

	/// Telling that the 3-bit BCID comes at "position 7" in the multiplicity word
	static const uint32_t MultiplicityBcidPos = 7;

	/// Weird mask for the bit showing if more than two muon candidates were in the trigger sector
	static const uint32_t CandidateOverflowMask = 0x1;

	/// Position of the candidate overflow mask
	static const uint32_t CandidateOverflowShift = 0;

	/// Weird mask for the bit showing if more than one muon candidates were in the sector RoI
	static const uint32_t RoIOverflowMask = 0x1;

	/// Position of the RoI overflow mask
	static const uint32_t RoIOverflowShift = 1;

	/// Bit in the candidate's address turned on for endcap candidates
	static const uint32_t EndcapAddressMask = 0x80;
	/// Bit in the candidate's address turned on for forward candidates
	static const uint32_t ForwardAddressMask = 0x40;

	/// Mask for extracting the sector ID for endcap candidates from the data word
	static const uint32_t EndcapSectorIdMask = 0x3f;
	/// Mask for extracting the sector ID for forward candidates from the data word
	static const uint32_t ForwardSectorIdMask = 0x1f;
	/// Mask for extracting the sector ID for barrel candidates from the data word
	static const uint32_t BarrelSectorIdMask = 0x1f;
	/// Mask for the bit showing which hemisphere the candidate came from.(1: positive; 0: negative)
	static const uint32_t SectorHemisphereMask = 0x1;

	/// Mask for extracting the RoI for barrel candidates from the data words
	static const uint32_t BarrelRoIMask = 0x1f;
	/// Mask for extracting the RoI for endcap candidates from the data words
	static const uint32_t EndcapRoIMask = 0xff;
	/// Mask for extracting the RoI for forward candidates from the data words
	static const uint32_t ForwardRoIMask = 0x3f;
	/// Position of the RoI bits in the data word
	static const uint32_t RoIShift = 2;

	/// Mask for extracting the overlap bits for barrel candidates from the data words
	static const uint32_t BarrelOverlapMask = 0x3;
	/// Position of the overlap bits in barrel data words
	static const uint32_t BarrelOverlapShift = 9;
	/// Mask for extracting the overlap bits for endcap candidates from the data words
	static const uint32_t EndcapOverlapMask = 0x1;
	/// Position of the overlap bits in endcap data words
	static const uint32_t EndcapOverlapShift = 10;

	/// Mask for extracting the p<sub>T</sub> threshold passed by the candidate from the data word
	static const uint32_t CandidatePtMask = 0x7;
	/// Position of the p<sub>T</sub> threshold bits in the data words
	static const uint32_t CandidatePtShift = 11;

	/// Mask for extracting the last 3 bits of the BCID of the muon candidate from the data word
	static const uint32_t CandidateBcidMask = 0x7;
	/// Position of the BCID bits in the data words
	static const uint32_t CandidateBcidShift = 14;

	/// Mask for extracting the address of the muon candidate from the data word
	static const uint32_t CandidateSectorAddressMask = 0xff;
	/// Position of the muon candidate's address in the data word
	static const uint32_t CandidateSectorAddressShift = 17;

	/// Mask for extracting the bit from the data word showing whether the candidate had the highest p<sub>T</sub> in the sector
	static const uint32_t CandidateHighestPtMask = 0x1;
	/// Position of the "highest p<sub>T</sub>" bit
	static const uint32_t CandidateHighestPtShift = 25;

	/// Mask for extracting the bit from the data word showing if the muon candidate was sent to the RoIB
	static const uint32_t CandidateSentRoIMask = 0x1;
	/// Position of the "candidate sent to RoIB" bit.
	static const uint32_t CandidateSentRoIShift = 26;

	/// Position of the bit turned on for the multiplicity words that distinguishes them from the data words
	static const uint32_t MultiplicityWordFlagShift = 29;

	/// Additional set of some decoder variables for RoIB words

	/// Sector address of the muon candidate from the RoI
	static const uint32_t RoIWordSectorAddressMask = 0xff;
	static const uint32_t RoIWordSectorAddressShift = 14;

	/// Pt of the muon candidate from the RoI
	static const uint32_t RoIWordPtMask = 0x7;
	static const uint32_t RoIWordPtShift = 11;

}

#endif // MUCTPI_DATAFORMAT_H
