## @package DQOksConfMaker.okswriter
# Module containing the tools to write a DQConfiguration in OKS format
# @version: 1
# @author: andrea.dotti@pi.infn.it
# 14 August 2008
"""
This module contains tools to write a DQConfiguration in the online DQMF (OKS XML) format
"""
from __future__ import print_function
from __future__ import absolute_import
from builtins import str
import DQOksConfMaker
from DQConfMakerBase.Helpers import BaseException
from DQConfMakerBase.DQElements import DQBase
from . import doctypeheader
from xml.dom.minidom import Document,Comment,DocumentType
from time import strftime
from socket import gethostname

## Writing exception
# OKS exception
class OksWritingError(BaseException):
    """
    This exception is thrown when an error adding and object to the OKS configuration is encountered
    """
    def __init__(self,obj):
        BaseException.__init__(self, "Could not convert to OKS XML Format the object:" + str(obj) )

## Generic exception
# OKS exception
class OksCannotCreateConf(BaseException):
    """
    This exception is thrown when an error creating the OKS configuration is encountered
    """
    def __init__(self,reason):
        BaseException.__init__(self, "Could not create OKS XML file ("+str(reason)+")" )

## Class document
# This class represents a valid OKS XML configuration document for DQMF
class OksDocument(Document):
    """
    This class represents the DQ configuration in the OKS XML format
    """
    ## Default constructor
    def __init__(self ):
        """
        Creates a OKS XML document
        """
        Document.__init__(self)
        self.appendChild( Comment('File auto generated by:'+self.__class__.__name__+' from module: '+__name__+' contact authors'))
        self.appendChild( DocumentType( doctypeheader.documenttype ) )
        self.numIncludes = 0
        self.numElements = 0
        self.root = self.createElement( 'oks-data' )
        self.includes = self.createElement( 'include' )
        self.root.appendChild( self.includes )
        self.info = self.createElement('info')
        self.root.appendChild( self.info)
        self.appendChild(self.root)
        self.finalized = False
    ## Finalize the document
    def finalize(self):
        """
        This methods fills some meta information of the document
        """
        if not self.finalized:
            self.finalized = True
            self.fillInfoElement()
    ## Writes to a file
    # @param filename: the name of the file to write the document
    def writeToFile(self,filename):
        """
        Writes the configuration to a file specified by name
        """
        self.finalize()
        f=open(filename,'w')
        f.write( self.toprettyxml(indent="  "))
        f.close()
    ## Fills metadata    
    def fillInfoElement(self):
        """
        Fills the info element of the OKS configuration
        """
        self.info.setAttribute('name','')
        self.info.setAttribute('type','')
        self.info.setAttribute('num-of-includes', str(self.numIncludes) )
        self.info.setAttribute('num-of-items', str(self.numElements) )
        self.info.setAttribute('oks-format', 'extended')
        self.info.setAttribute('oks-version','oks-03-06-02 built &quot;Jan 24 2007&quot')
        self.info.setAttribute('created-by','adotti')
        self.info.setAttribute('created-on', 'mypc' )
        self.info.setAttribute('creation-time', '14 Aug 2008 00:00:00' )
        self.info.setAttribute('last-modified-by', self.__class__.__name__)
        self.info.setAttribute('last-modified-on', gethostname() )
        self.info.setAttribute('last-modification-time', strftime("%Y-%m-%d %H:%M:%S"))
    ## Adds an include statement    
    # @param includepath: the included file string
    def addInclude(self,includepath):
        """
        Add an include file statement
        """
        include = self.createElement('file')
        include.setAttribute('path',includepath)
        self.includes.appendChild(include)
        self.numIncludes += 1
    ## Adds a DQBase object
    # @param object: the object to add. The object must be a valid DQBase object
    # @raise OksWritingError: in case of errors.
    def addObject(self , object ):
        """
        Add an DQElement object to OKS document.
        """
        self.numElements += 1
        try:
            element = self.createElement( object.type )
            element.setAttribute('id', object.id )
            for key, attribute in object.attr.items():
                att = self.createElement('attr')
                element.appendChild(att)
                att.setAttribute('name', key )
                att.setAttribute('type', attribute[0])
                if type(attribute[1])==type([]): # Multi value attribute
                    att.setAttribute('num',str(len(attribute[1])))
                tostring=str(attribute[1]).strip("[]").replace(',',' ').replace("'",' ')
                if tostring != "":
                    v = self.createTextNode( tostring )
                    att.appendChild(v)
            for key,relation in object.rels.items():
                rel = self.createElement( 'rel' )
                element.appendChild( rel )
                rel.setAttribute('name', key )
                # handle special case for Algorithm, it's not a is-multi-value
                if key != 'Algorithm':
                    #relations have always at least one element: None
                    rel.setAttribute('num', str(len(relation)-1) )
                for r in relation:
                    if not r: #skip empty relations
                        continue
                    mystring = str(r.type)+" "+str(r.id)
                    if len(relation)>1:
                        mystring += "\n"
                    v = self.createTextNode(mystring)
                    rel.appendChild( v )   
            self.root.appendChild( element )
        except:
            raise OksWritingError(object)

## Writes the configuration to file
# @param filename: the file name to which write the configuration
# @param configList: a list of valid dq configuraiton elements. A list of valid DQBase objects
# @param useTDAQalgoDef: if True use DQAlgorithm definition from tdaq databases, else use the definitions in configList
# @raise OksCannotCreateConf: if failed creating the OKS XML file
# @note: If the configuration file has to be used online set useTDAQalgoDef to True
def writeOksConfiguration( filename = 'dq.data.xml' , configList= [] , useTDAQalgoDefinitions = True):
    """
    Write the dq configuration into OKS XML format to a file.
    The input configuration is a list of DQElements.
    If useTDAQalgoDefinitions is False generate also OKS representation of the DQAlgorithms. In normal condition this should not be used.
    """
    numElements = 0
    try:
        doc = OksDocument()
        if useTDAQalgoDefinitions:
            doc.addInclude( "daq/sw/dqm-algorithms.data.xml" )
        for c in configList:
            if not (useTDAQalgoDefinitions and c.type == "DQAlgorithm"):
                if DQOksConfMaker._isAcceptedOksElement(c):
                    doc.addObject( c )
                    numElements += 1
                else:
                    print("==========WARNING===========")
                    print("Object: "+str(c)+" does not have a valid OKS representation, cannot save")
                    print("The object is referenced by:",c.isReferencedBy)
                    print("You may need to manually edit the configuration")
                    print("============================")
        doc.writeToFile(filename)
    except OksWritingError as msg:
        print("Failed creating file: "+str(filename)+": "+str(msg))
        raise OksCannotCreateConf( msg )
    except Exception as msg:
        print("Failed creating file: "+str(filename)+": "+str(msg))
        raise OksCannotCreateConf( msg )
    print("Wrote: "+str( numElements )+" elements to: "+str(filename))
        
