//Dear emacs, this is -*- c++ -*-

/**
 * @file src/old/v30_FullEventFragment.h
 * @author <a href="mailto:Andre.dos.Anjos@cern.ch">Andre DOS ANJOS</a> 
 * $Author$
 * $Revision$
 * $Date$
 *
 * @brief Defines the Event Fragment entity. The definition is based on the
 * update of ATL-DAQ-98-129, by D.Francis et al.
 */

#ifndef EFORMAT_V30_FULLEVENTFRAGMENT_H
#define EFORMAT_V30_FULLEVENTFRAGMENT_H

#include "src/old/v31_Header.h"

namespace eformat {

  namespace v30 {

    /**
     * Describes how to access the contents of an event fragment, as prescribed
     * by the event format note.
     */
    class FullEventFragment : public eformat::v31::Header {

    public: //interface

      /**
       * To build a fragment given the containing buffer. I need to know
       * where the fragment starts in order to do that.
       *
       * @param it The exact position where this fragment should start.
       */
      FullEventFragment (const uint32_t* it);

      /**
       * Builds an empty, otherwise useless FullEventFragment
       */
      FullEventFragment () : eformat::v31::Header(), m_start() {}

      /**
       * Copy constructor
       *
       * @param other The fragment to be copied
       */
      FullEventFragment (const FullEventFragment& other) 
        : eformat::v31::Header(other), m_start(other.m_start) {}

      /**
       * Destructor virtualisation
       */
      virtual ~FullEventFragment() {}

      /**
       * Copy operator
       *
       * @param other The fragment to be copied
       */
      FullEventFragment& operator= (const FullEventFragment& other) 
      { 
        eformat::v31::Header::operator=(other); 
        m_start=other.m_start; 
        return *this; 
      }

      /**
       * Manual re-assignment
       *
       * @param it The position pointing the first word of this fragment
       */
      FullEventFragment& assign (const uint32_t* it);

      /**
       * Says if the the fragment is valid. This may throw exceptions.
       */
      virtual bool check (const uint16_t to_check=eformat::MAJOR_V30_VERSION) const;
      
      /**
       * Says if the Fragment and all children are valid.
       */
      bool check_tree (const uint16_t to_check=eformat::MAJOR_V30_VERSION) const;
      
      /**
       * Returns the date, in seconds elapsed since the 1st. January,
       * 1970. The timezone is fixed to UTC.
       */
      inline uint32_t date() const
      { return m_start[0]; }
      
      /**
       * Returns the global identifier
       */
      inline uint32_t global_id() const
      { return m_start[1]; }
      
      /**
       * Returns the run number
       */
      inline uint32_t run_no() const
      { return m_start[2]; }
      
      /**
       * Returns the lvl1 identifier
       */
      inline uint32_t lvl1_id() const
      { return m_start[3]; }
      
      /**
       * Retursn the lvl1 trigger type
       */
      inline uint32_t lvl1_trigger_type() const
      { return m_start[4]; }
      
      /**
       * Returns the lvl2 trigger info
       */
      inline uint32_t lvl2_trigger_info() const
      { return m_start[5]; }
      
      /**
       * Returns the number of Event Filter words (constant = 4)
       */
      inline uint32_t nevent_filter_info () const { return 4; }
      
      /**
       * Returns an iterator to the first of the (4) EF words
       */
      inline const uint32_t* event_filter_info(void) const 
      { return m_start + 6; }
      
    private: //static stuff
      
      static const uint32_t NSPECIFIC;

    private: //representation
      
      const uint32_t* m_start; ///< my start word

    };

  }

}

#endif /* EFORMAT_V30_FULLEVENTFRAGMENT_H */
